"""Implement SeshetBot as subclass of ircutils3.bot.SimpleBot."""

import os
import sys
import traceback
from io import StringIO
from datetime import datetime

from ircutils3 import bot, client

from .utils import KVStore, Storage


class SeshetBot(bot.SimpleBot):
    """Extend `ircutils3.bot.SimpleBot`.
    
    Each instance represents one bot, connected to one IRC network.
    Each instance should have its own database, but can make use of
    any shared command modules. The modules may have to be added to
    the bot's database if the bot wasn't created using the
    `seshet --config` or `seshet --new` commands.
    """
    
    session = Storage()
    """Shared runtime storage available for all command modules."""
    
    storage = None
    """If bot is initialized with a database connection, persistent
    KV store available for all command modules. Each module will have
    its own namespace.
    """
    
    def __init__(self, nick='Seshet', db=None):
        """Extend `ircutils3.bot.SimpleBot.__init__()`.
        
        Keyword argument `db` is required for running commands other
        than core commands and should be an instance of pydal.DAL.
        """
        
        bot.SimpleBot.__init__(self, nick, auto_handle=False)
        
        if db is None:
            # no database connection, only log to file and run
            # core command modules
            self.log = self._log_to_file
            self.run_commands = self._run_only_core
            
            # dummy KV store since no db
            self.storage = Storage()
        else:
            self.db = db
            self.storage = KVStore(db)
        
        # Add default handlers
        self.events["any"].add_handler(client._update_client_info)
        self.events["ctcp_version"].add_handler(client._reply_to_ctcp_version)
        
    def log(self, etype, source, msg='', target='', hostmask='', parms=''):
        """Log an event in the database.
        
        Required:
            `etype` - event type. One of 'PRIVMSG', 'QUIT', 'PART', 'ACTION',
                'NICK', 'JOIN', 'MODE', 'KICK', 'CTCP', or 'ERROR'. Enforced 
                by database model.
            `source` - source of the event. Usually a user. For NICK events,
                the user's original nickname. For ERROR events, this should be
                the exception name, the module name, and the line number. The
                full traceback will be logged in `msg`.
        Optional:
            `msg` - a message associated with the event.
            `target` - the target the message was directed to. For MODE and KICK
                events, this will be the user the event was performed on. For
                NICK events, this will be channel the event was seen in (an event
                will be created for each channel the user is seen by the bot in).
            `hostmask` - a hostmask associated with the event.
            `parms` - any additional parameters associated with the event, such as
                a new nickname (for NICK events), mode switches (for MODE events),
                or a dump of local variables (for ERROR events).
        """
        
        self.db.event_log.insert(event_type=etype,
                                 event_time=datetime.utcnow(),
                                 source=source,
                                 target=target,
                                 message=msg,
                                 host=hostmask,
                                 parms=parms,
                                 )
        self.db.commit()
        
    def run_commands(self, e):
        pass
    
    def on_message(self, e):
        self.log('PRIVMSG', e.source, e.message, e.target)
        self.run_commands(e)
    
    def on_join(self, e):
        self.log('JOIN', e.source, '', e.target, e.user+'@'+e.host)
        self.run_commands(e)
    
    def on_part(self, e):
        pass
    
    def on_quit(self, e):
        pass
    
    def on_disconnect(self, e):
        pass
    
    def on_kick(self, e):
        pass
    
    def on_nick_change(self, e):
        pass
    
    def on_ctcp_action(self, e):
        pass
    
    def on_welcome(self, e):
        pass
    
    def on_mode(self, e):
        # will be changed
        self.log('MODE', e.source, str(e.params), e.target)
    
    def before_poll(self):
        """Called each loop before polling sockets for I/O."""
        pass
    
    def after_poll(self):
        """Called each loop after polling sockets for I/O and
        handling any queued events.
        """
        pass
    
    def connect(self, *args, **kwargs):
        """Extend `client.SimpleClient.connect()` with defaults"""
        defaults = {}

        for i, k in enumerate(('host', 'port', 'channel', 'use_ssl', 'password')):
            if i < len(args):
                defaults[k] = args[i]
            elif k in kwargs:
                defaults[k] = kwargs[k]
            else:
                def_k = 'default_' + k
                defaults[k] = getattr(self, def_k, None)

        if defaults['use_ssl'] is None:
            defaults['use_ssl'] = False

        if defaults['host'] is None:
            raise TypeError("missing 1 required positional argument: 'host'")

        client.SimpleClient.connect(self, **defaults)

    def start(self):
        self._loop(self.conn._map)
    
    def _remove_user(self, e):
        pass
    
    def _log_to_file(self, etype, source, msg='', target='', hostmask='', parms=''):
        """Override `log()` if bot is not initialized with a database
        connection. Do not call this method directly.
        """
        if etype == 'PRIVMSG':
            log = open('seshet.log', 'a')
            log.write("<{}> {}\n".format(source, msg))
            log.close()
        elif etype == 'MODE':
            log = open('seshet.log', 'a')
            log.write("{} MODE {} {}\n".format(source, target, msg))
            log.close()
    
    def _run_only_core(self, *args, **kwargs):
        """Override `_run_commands()` if bot is not initialized with a
        database connection. Do not call this method directly.
        
        Rather than getting a list of enabled modules from the database,
        Seshet will only run the commands defined by `core` in this package.
        The bot will only run commands given in private message ("query")
        by either an authenticated user defined in the instance's config file,
        or by any user with the same hostmask if authentication isn't set up
        in the instance's config file.
        
        The `core` command module from this package can only be overridden if
        the bot is initialized with a database connection and a new `core`
        module is entered into the database.
        """
        pass

    def _loop(self, map):
        """The main loop. Poll sockets for I/O and run any other functions
        that need to be run every loop.
        """
        try:
            from asyncore import poll
        except ImportError:
            raise Exception("Couldn't find poll function. Cannot start bot.")
    
        while map:
            self.before_poll()
            poll(timeout=30.0, map=map)
            self.after_poll()



"""Various utility classes used by the bot and command modules.

`Storage` is copied from gluon.storage, part of the web2py framework,
    Copyrighted by Massimo Di Pierro <mdipierro@cs.depaul.edu>,
    License: LGPLv3 (http://www.gnu.org/licenses/lgpl.html)
"""

import random
import inspect
import pickle

from pydal import Field


class Storage(dict):
    """A Storage object is like a dictionary except `obj.foo` can be used
    in addition to `obj['foo']`, and setting obj.foo = None deletes item foo.

    Example:

        >>> o = Storage(a=1)
        >>> print o.a
        1

        >>> o['a']
        1

        >>> o.a = 2
        >>> print o['a']
        2

        >>> del o.a
        >>> print o.a
        None

    """
    
    __slots__ = ()
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    __getitem__ = dict.get
    __getattr__ = dict.get
    __getnewargs__ = lambda self: getattr(dict,self).__getnewargs__(self)
    __repr__ = lambda self: '<Storage %s>' % dict.__repr__(self)
    __getstate__ = lambda self: None
    __copy__ = lambda self: Storage(self)

    def getlist(self, key):
        """Returns a Storage value as a list.

        If the value is a list it will be returned as-is.
        If object is None, an empty list will be returned.
        Otherwise, `[value]` will be returned.

        Example output for a query string of `?x=abc&y=abc&y=def`::

            >>> request = Storage()
            >>> request.vars = Storage()
            >>> request.vars.x = 'abc'
            >>> request.vars.y = ['abc', 'def']
            >>> request.vars.getlist('x')
            ['abc']
            >>> request.vars.getlist('y')
            ['abc', 'def']
            >>> request.vars.getlist('z')
            []

        """
        
        value = self.get(key, [])
        if value is None or isinstance(value, (list, tuple)):
            return value
        else:
            return [value]

    def getfirst(self, key, default=None):
        """Returns the first value of a list or the value itself when given a
        `request.vars` style key.

        If the value is a list, its first item will be returned;
        otherwise, the value will be returned as-is.

        Example output for a query string of `?x=abc&y=abc&y=def`::

            >>> request = Storage()
            >>> request.vars = Storage()
            >>> request.vars.x = 'abc'
            >>> request.vars.y = ['abc', 'def']
            >>> request.vars.getfirst('x')
            'abc'
            >>> request.vars.getfirst('y')
            'abc'
            >>> request.vars.getfirst('z')

        """
        
        values = self.getlist(key)
        return values[0] if values else default

    def getlast(self, key, default=None):
        """Returns the last value of a list or value itself when given a
        `request.vars` style key.

        If the value is a list, the last item will be returned;
        otherwise, the value will be returned as-is.

        Simulated output with a query string of `?x=abc&y=abc&y=def`::

            >>> request = Storage()
            >>> request.vars = Storage()
            >>> request.vars.x = 'abc'
            >>> request.vars.y = ['abc', 'def']
            >>> request.vars.getlast('x')
            'abc'
            >>> request.vars.getlast('y')
            'def'
            >>> request.vars.getlast('z')

        """
        
        values = self.getlist(key)
        return values[-1] if values else default


class KVStore:
    """Create a key/value store in the bot's database for each
    command module to use for persistent storage. Can be accessed
    either like a class:
    
        >>> store = KVStore(db)
        >>> store.foo = 'bar'
        >>> store.foo
        'bar'
        
    Or like a dict:
    
        >>> store[spam] = 'eggs'
        >>> store[spam]
        'eggs'
        
    The KVStore object uses `inspect` to determine which module
    the object is being accessed from and will automatically create
    a database table as needed or determine which one to use if it
    already exists, so that each module the object is used from has
    its own namespace.
    
    KVStore has most of the same interfaces as an ordinary `dict`, but
    is not a subclass of `dict` or `collections.UserDict` because
    so many functions had to be completely rewritten to work with
    KVStore's database model.
    """

    def __init__(self, db):
        # make sure some tables are defined:
        
        if 'modules' not in db:
            # list of registered modules
            db.define_table('modules', Field('name'))
            
        for m in db().select(db.modules.ALL):
            # these are modules' own "namespaces"
            tbl_name = 'kv_' + m.name
            if tbl_name not in db:
                db.define_table(tbl_name,
                                Field('k', 'string', unique=True),
                                Field('v', 'text'),
                                )
        
        self._db = db   # pydal DAL instance
        # It's recommended to use a separate database
        # for the bot and for the KV store to avoid
        # accidental or malicious name collisions
        #
        # (Then why doesn't the default implimentation?)
        
    def __getattr__(self, k):
        if k.startswith('_'):
            return self.__dict__[k]
        
        db = self._db
        
        tbl = self._get_calling_module()
        tbl_name = 'kv_' + tbl if table is not None else None
        if tbl is None or tbl_name not in db:
            # table doesn't exist
            return None
        
        r = db(db[tbl_name].k == k)
        if r.isempty():
            # no db entry for this key
            return None
        
        r = r.select().first()
        
        # db should return string, pickle expects bytes
        return pickle.loads(r.v.encode(errors='ignore'))

    def __setattr__(self, k, v):
        if k.startswith('_'):
            self.__dict__[k] = v
            return
        elif k in self.__dict__:
            # instance attributes should be read-only-ish
            raise AttributeError("Name already in use: %s" % k)
        
        db = self._db
        
        if v is not None:
            v = pickle.dumps(v).decode(errors='ignore')
        
        tbl = self._get_calling_module()
        tbl_name = 'kv_' + tbl if tbl is not None else None
        
        if tbl is None or tbl_name not in db:
            if v is not None:
                # module not registered, need to create
                # a new table
                self._register_module(tbl)
                db[tbl_name].insert(k=k, v=repr(v))
            else:
                # no need to delete a non-existent key
                return None
        else:
            if v is not None:
                db[tbl_name].update_or_insert(db[tbl_name].k == k, k=k, v=v)
            else:
                db(db[tbl_name].k == k).delete()
        
        db.commit()
        self._db = db
    
    def __delattr__(self, k):
        self.__setattr__(k, None)
    
    def __getitem__(self, k):
        return self.__getattr__(k)
    
    def __setitem__(self, k, v):
        self.__setattr__(k, v)
    
    def __delitem__(self, k):
        self.__setattr__(k, None)

    def _register_module(self, name):
        db = self._db
        
        #if name == 'modules':
            # name collision with existing table
            #raise ValueError('The module name %s is reserved' % name)
        tbl_name = 'kv_' + name
            
        if db(db.modules.name == name).isempty():
            db.modules.insert(name=name)
            db.commit()
        if tbl_name not in db:
            db.define_table(tbl_name,
                            Field('k', 'string', unique=True),
                            Field('v', 'text'),
                            )
        self._db = db

    def _get_calling_module(self):
        # in theory, bot modules will be registered with register_module
        # when they're uploaded and installed

        curfrm = inspect.currentframe()
        for f in inspect.getouterframes(curfrm)[1:]:
            if self.__module__.split('.')[-1] not in f[1]:
                calling_file = f[1]
                break
        caller_mod = inspect.getmodulename(calling_file)
    
        db = self._db
        mod = db(db.modules.name == caller_mod)
        if mod.isempty():
            return None
        else:
            return caller_mod
        
    def keys(self):
        db = self._db
        tbl = self._get_calling_module()
        tbl_name = 'kv_' + tbl if tbl is not None else None
        if tbl is None or tbl_name not in db:
            return []
        all_items = db().select(db[tbl_name].ALL)
        all_keys = [r.k for r in all_items]
        return all_keys
    
    def values(self):
        all_keys = self.keys()
        all_vals = list()
        for k in all_keys:
            all_vals.append(self[k])
        return all_vals
    
    def update(self, other):
        for k, v in other.items():
            self[k] = v
        return None
    
    def items(self):
        return zip(self.keys(), self.values())
    
    def iterkeys(self):
        return iter(self.keys())
    
    def itervalues(self):
        return iter(self.values())
    
    def iteritems(self):
        return iter(self.items())
    
    def __iter__(self):
        return iter(self.keys())
    
    def __contains__(self, k):
        if self[k] is not None:
            return True
        else:
            return False
        
    def __copy__(self):
        """Return a dict representing the current table"""
        d = dict()
        d.update(self.items())
        return d
    
    def copy(self):
        """Return a dict representing the current table"""
        return self.__copy__()
    
    def pop(self, k):
        v = self[k]
        self[k] = None
        return v
    
    def popitem(self):
        """Unlike `dict.popitem()`, this is actually random"""
        all_items = self.items()
        removed_item = random.choice(all_items)
        self[removed_item[0]] = None
        return removed_item
    
    def setdefault(self, k, v=None):
        existing_v = self[k]
        if existing_v is None:
            self[k] = v
            return v
        return existing_v
    
    def has_key(self, k):
        return k in self
    
    def get(self, k, v=None):
        existing_v = self[k]
        if existing_v is None:
            return v
        else:
            return existing_v
        
    def clear(self):
        for k in self.keys():
            self[k] = None



"""Define default configuration, read configuration file, and apply
configuration to SeshetBot instance.
"""

from configparser import ConfigParser

from pydal import DAL, Field


default_config = """
[connection]
# passed to SeshetBot.connect()
server: chat.freenode.net
port: 6667
channels: #botwar
ssl: False

[client]
nickname: Seshet
user: seshet
realname: seshetbot

[welcome]
# stuff sent by the bot after connecting
use_nickserv: False
nickserv_pass:
user_mode: -x

[locale]
timezone: UTC
locale: en_US
# see docs for datetime.strftime
date_fmt: %m%d%y
# 071415
time_fmt: %H:%M:%S
# 11:49:57
short_datetime_fmt: %Y-%m-%d %H:%M:%S
# 2015-07-14 11:49:57
long_datetime_fmt: %A %d %B %Y at %H:%M:%S %Z
# Tuesday 14 July 2015 at 11:49:57 UTC

[database]
use_db: True
db_string: sqlite://seshet.db

[logging]
# if using db, this will be ignored
file: logs/%(target)s_%(date)s.log

[debug]
# corresponds to levels in logging module or None
verbosity: warning
file: seshet-debug.log
"""


testing_config = """
[connection]
# passed to SeshetBot.connect()
server: chat.freenode.net
port: 6667
channels: #botwar
ssl: False

[client]
nickname: Seshet
user: seshet
realname: seshetbot

[welcome]
# stuff sent by the bot after connecting
use_nickserv: False
nickserv_pass:
user_mode: -x

[locale]
timezone: UTC
locale: en_US
# see docs for datetime.strftime
date_fmt: %m%d%y
# 071415
time_fmt: %H:%M:%S
# 11:49:57
short_datetime_fmt: %Y-%m-%d %H:%M:%S
# 2015-07-14 11:49:57
long_datetime_fmt: %A %d %B %Y at %H:%M:%S %Z
# Tuesday 14 July 2015 at 11:49:57 UTC

[database]
# no db connection for testing
use_db: False

[logging]
# if using db, this will be ignored
file: logs/%(target)s_%(date)s.log

[debug]
# corresponds to levels in logging module or None
verbosity: debug
file: seshet-debug.log
"""


def build_db_tables(db):
    """Build Seshet's basic database schema. Requires one parameter,
    `db` as `pydal.DAL` instance.
    """
    if not isinstance(db, DAL) or not db._uri:
        raise Exception("Need valid DAL object to define tables")

    # event log - self-explanatory, logs all events
    db.define_table('event_log',
                    Field('event_type'),
                    Field('event_time', 'datetime'),
                    Field('source'),
                    Field('target'),
                    Field('message', 'text'),
                    Field('host'),
                    Field('parms', 'list:string'),
                    )
        

def build_bot(config_file=None):
    """Parse a config and return a SeshetBot instance. After, the bot can be run
    simply by calling .connect() and then .start()
    
    Optional arguments:
        config_file - valid file path or ConfigParser instance
        
        If config_file is None, will read default config defined in this module.
    """
    from . import bot

    config = ConfigParser()
    if config_file is None:
        config.read_string(default_config)
    elif isinstance(config_file, ConfigParser):
        config = config_file
    else:
        config.read(config_file)

    # shorter names
    db_conf = config['database']
    conn_conf = config['connection']
    client_conf = config['client']
    # add more as they're used

    if db_conf.getboolean('use_db'):
        db = DAL(db_conf['db_string'])
        build_db_tables(db)
    else:
        db = None

    seshetbot = bot.SeshetBot(client_conf['nickname'], db)

    # connection info for connect()
    seshetbot.default_host = conn_conf['server']
    seshetbot.default_port = int(conn_conf['port'])
    seshetbot.default_channel = conn_conf['channels'].split(',')
    seshetbot.default_use_ssl = conn_conf.getboolean('ssl')

    # client info
    seshetbot.user = client_conf['user']
    seshetbot.real_name = client_conf['realname']

    return seshetbot


#!/usr/bin/env python3

"""Seshet - Modular, dynamic IRC bot"""

from setuptools import setup
from os import path


long_description = open(
    path.join(
        path.dirname(__file__),
        'README.rst'
    )
).read()

setup(
    name='seshet',
    version='0.1.0-2',
    url='https://github.com/kopachris/seshet',
    license='BSD',
    author='Christopher Koch',
    author_email='ch_koch@outlook.com',
    description='Modular, dynamic IRC bot',
    long_description=long_description,
    packages=['seshet'],
    scripts = ['test.py'],
    install_requires=[
        'ircutils3',
        'pydal',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Operating System :: OS Independent',

        'License :: OSI Approved :: BSD License',

        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',

        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Topic :: Software Development :: Libraries :: Application Frameworks',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



#!/usr/bin/env python3

from seshet import bot

nick = input("Nickname? (Seshet) ") or "Seshet"
server = input("Server? (chat.freenode.net) ") or "chat.freenode.net"
port = input("Port? (6667) ") or 6667
channel = input("Channel to join? (#botwar) ") or "#botwar"

seshetbot = bot.SeshetBot(nick)

print("Connecting...", end=" ")
seshetbot.connect(server, port=int(port), channel=channel)
print("Done.\nPress Ctrl-C to quit.")
seshetbot.start()
